@@ -27,17 +27,17 @@ use std::path::{Path, PathBuf};
 use std::str::FromStr;
 use std::{fmt, fs, io};
 
-use amplify::hex::{FromHex, ToHex};
+use amplify::hex::ToHex;
 use amplify::{IoError, Wrapper};
-use bitcoin::consensus::{Decodable, Encodable};
+use bitcoin::psbt::serialize::Serialize;
+use bitcoin::psbt::PartiallySignedTransaction;
 use bitcoin::secp256k1::Secp256k1;
 use bitcoin::util::address;
 use bitcoin::util::bip32::{ChildNumber, ExtendedPubKey};
-use bitcoin::util::psbt::raw::ProprietaryKey;
-use bitcoin::util::psbt::{PartiallySignedTransaction as Psbt, PsbtParseError};
-use bitcoin::{Address, Network};
+use bitcoin::{consensus, Address, Network};
 use bitcoin_hd::DeriveError;
 use bitcoin_onchain::UtxoResolverError;
+use bitcoin_scripts::taproot::DfsPath;
 use bitcoin_scripts::PubkeyScript;
 use clap::Parser;
 use colored::Colorize;
@@ -46,6 +46,8 @@ use electrum_client::ElectrumApi;
 use miniscript::psbt::PsbtExt;
 use miniscript::{Descriptor, MiniscriptKey, TranslatePk};
 use psbt::construct::{self, Construct};
+use psbt::serialize::Deserialize;
+use psbt::{ProprietaryKeyDescriptor, ProprietaryKeyError, ProprietaryKeyLocation};
 use slip132::{
     DefaultResolver, FromSlip132, KeyApplication, KeyVersion, ToSlip132, VersionResolver,
 };
@@ -54,6 +56,7 @@ use wallet::descriptors::InputDescriptor;
 use wallet::hd::{Descriptor as DescrTrait, SegmentIndexes, TrackingAccount, UnhardenedIndex};
 use wallet::locks::LockTime;
 use wallet::onchain::ResolveUtxo;
+use wallet::psbt::{Psbt, PsbtParseError};
 
 /// Command-line arguments
 #[derive(Parser)]
@@ -73,7 +76,7 @@ pub struct Args {
     ///
     /// Used only by `check`, `history`, `construct` and some forms of
     /// `extract` command
-    #[clap(short, long, global = true, default_value("pandora.network"))]
+    #[clap(short, long, global = true, default_value("electrum.blockstream.info"))]
     pub electrum_server: String,
 
     /// Customize electrum server port number. By default the wallet will use
@@ -204,10 +207,11 @@ specified at the end of input descriptor.
 
 Sequence number representations:
 - `rbf(SEQ)`: use replace-by-fee opt-in for this input;
-- `after(NO)`: allow the transaction to be mined with sequence lock
-  to `NO` blocks;
-- `older(NO)`: allow the transaction to be mined if it is older then
-  the provided number `NO` of 5-minute intervals.
+- `height(NO)`: allow the transaction to be mined with sequence lock
+  set to `NO` blocks (required for miniscript `older` satisfaction);
+- `time(NO)`: allow the transaction to be mined if it is older then
+  the provided number `NO` of 5-minute intervals (required for miniscript 
+  `after` satisfaction).
 
 SIGHASH_TYPE representations:
 - `ALL` (default)
@@ -232,8 +236,15 @@ SIGHASH_TYPE representations:
         #[clap(short, long, default_value = "0")]
         change_index: UnhardenedIndex,
 
-        #[clap(long = "proprietary-key")]
-        propriatary_keys: Vec<ProprietaryKeyDescriptor>,
+        /// Allows adding different forms of commitments to the change output,
+        /// if it is present.
+        #[clap(long)]
+        allow_tapret_path: Option<DfsPath>,
+
+        /// Additional proprietary keys which will be added to the constructed
+        /// PSBT.
+        #[clap(short = 'k', long = "proprietary-key")]
+        proprietary_keys: Vec<ProprietaryKeyDescriptor>,
 
         /// Destination file to save constructed PSBT
         psbt_file: PathBuf,
@@ -247,9 +258,6 @@ SIGHASH_TYPE representations:
 
     /// Try to finalize PSBT
     Finalize {
-        /// File containing fully-signed PSBT
-        psbt_file: PathBuf,
-
         /// Destination file to save binary transaction. If no file is given
         /// the transaction is print to the screen in hex form.
         #[clap(short = 'o', long = "output")]
@@ -259,6 +267,9 @@ SIGHASH_TYPE representations:
         /// to specify some custom network (testnet, for instance).
         #[clap(long)]
         publish: Option<Option<Network>>,
+
+        /// File containing fully-signed PSBT
+        psbt_file: PathBuf,
     },
 
     /// Get info about extended public key data
@@ -320,7 +331,8 @@ impl Args {
                 inputs,
                 outputs,
                 change_index,
-                propriatary_keys,
+                proprietary_keys,
+                allow_tapret_path,
                 psbt_file,
                 fee,
             } => self.construct(
@@ -329,7 +341,8 @@ impl Args {
                 inputs,
                 outputs,
                 *change_index,
-                propriatary_keys,
+                allow_tapret_path.as_ref(),
+                proprietary_keys,
                 *fee,
                 psbt_file,
             ),
@@ -570,7 +583,8 @@ impl Args {
         inputs: &[InputDescriptor],
         outputs: &[AddressAmount],
         change_index: UnhardenedIndex,
-        propriatary_keys: &Vec<ProprietaryKeyDescriptor>,
+        allow_tapret_path: Option<&DfsPath>,
+        proprietary_keys: &Vec<ProprietaryKeyDescriptor>,
         fee: u64,
         psbt_path: &Path,
     ) -> Result<(), Error> {
@@ -592,6 +606,11 @@ impl Args {
             "\nWallet descriptor:".bright_white(),
             descriptor
         );
+
+        if !matches!(descriptor, Descriptor::Tr(_)) && allow_tapret_path.is_some() {
+            return Err(Error::TapretRequiresTaproot);
+        }
+
         eprint!(
             "Re-scanning network {} using {} ... ",
             network.to_string().yellow(),
@@ -627,7 +646,15 @@ impl Args {
             &tx_map,
         )?;
 
-        for key in propriatary_keys {
+        if let Some(tapret_path) = allow_tapret_path {
+            for output in &mut psbt.outputs {
+                if !output.bip32_derivation.is_empty() {
+                    output.set_tapret_dfs_path(tapret_path);
+                }
+            }
+        }
+
+        for key in proprietary_keys {
             match key.location {
                 ProprietaryKeyLocation::Input(pos) if pos as usize >= psbt.inputs.len() => {
                     return Err(ProprietaryKeyError::InputOutOfRange(pos, psbt.inputs.len()).into())
@@ -654,8 +681,7 @@ impl Args {
             }
         }
 
-        let file = fs::File::create(psbt_path)?;
-        psbt.consensus_encode(file)?;
+        fs::write(psbt_path, &psbt.serialize())?;
 
         println!("{} {}\n", "PSBT:".bright_white(), psbt);
 
@@ -670,8 +696,8 @@ impl Args {
     ) -> Result<(), Error> {
         let secp = Secp256k1::new();
 
-        let file = fs::File::open(psbt_path)?;
-        let mut psbt = Psbt::consensus_decode(&file)?;
+        let data = fs::read(psbt_path)?;
+        let mut psbt = consensus::encode::deserialize::<PartiallySignedTransaction>(&data)?;
 
         psbt.finalize_mut(&secp).map_err(VecDisplay::from)?;
 
@@ -705,8 +731,8 @@ impl Args {
 
     fn inspect(&self, path: Option<&PathBuf>) -> Result<(), Error> {
         let psbt = if let Some(path) = path {
-            let file = fs::File::open(path)?;
-            Psbt::consensus_decode(&file)?
+            let data = fs::read(path)?;
+            Psbt::deserialize(&data)?
         } else {
             eprint!("Type in Base58 encoded PSBT and press enter: ");
             stdout().flush()?;
@@ -719,8 +745,8 @@ impl Args {
     }
 
     fn convert(&self, path: &Path) -> Result<(), Error> {
-        let file = fs::File::open(path)?;
-        let psbt = Psbt::consensus_decode(&file)?;
+        let data = fs::read(path)?;
+        let psbt = Psbt::deserialize(&data)?;
         println!("\n{}\n", psbt);
         Ok(())
     }
@@ -781,176 +807,6 @@ impl FromStr for AddressAmount {
     }
 }
 
-#[derive(Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Debug, Display, Error)]
-#[display(doc_comments)]
-pub enum ProprietaryKeyError {
-    /// incorrect proprietary key location `{0}`; allowed location formats are
-    /// `input(X)`, `output(X)` and `global`, where `X` is a 16-bit decimal
-    /// integer.
-    WrongLocation(String),
-
-    /// incorrect proprietary key type definition `{0}`.
-    ///
-    /// Type definition must start with a ket prefix in form of a short ASCII
-    /// string, followed by a key subtype represented by a 8-bit decimal integer
-    /// in parentheses without whitespacing. Example: `DBC(5)`
-    WrongType(String),
-
-    /// incorrect proprietary key format `{0}`.
-    ///
-    /// Proprietary key descriptor must consists of whitespace-separated three
-    /// parts:
-    /// 1) key location, in form of `input(no)`, `output(no)`, or `global`;
-    /// 2) key type, in form of `prefix(no)`;
-    /// 3) key-value pair, in form of `key:value`, where both key and value
-    ///    must be hexadecimal bytestrings; one of them may be omitted
-    ///    (for instance, `:value` or `key:`).
-    ///
-    /// If the proprietary key does not have associated data, the third part of
-    /// the descriptor must be fully omitted.
-    WrongFormat(String),
-
-    /// input at index {0} exceeds the number of inputs {1}
-    InputOutOfRange(u16, usize),
-
-    /// output at index {0} exceeds the number of outputs {1}
-    OutputOutOfRange(u16, usize),
-}
-
-#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Debug, Display)]
-pub enum ProprietaryKeyLocation {
-    #[display("global")]
-    Global,
-
-    #[display("input({0})")]
-    Input(u16),
-
-    #[display("output({0})")]
-    Output(u16),
-}
-
-impl FromStr for ProprietaryKeyLocation {
-    type Err = ProprietaryKeyError;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        let mut split = s.trim_end_matches(')').split('(');
-        match (
-            split.next(),
-            split.next().map(u16::from_str).transpose().ok().flatten(),
-            split.next(),
-        ) {
-            (Some("global"), None, _) => Ok(ProprietaryKeyLocation::Global),
-            (Some("input"), Some(pos), None) => Ok(ProprietaryKeyLocation::Input(pos)),
-            (Some("output"), Some(pos), None) => Ok(ProprietaryKeyLocation::Output(pos)),
-            _ => Err(ProprietaryKeyError::WrongLocation(s.to_owned())),
-        }
-    }
-}
-
-#[derive(Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Debug, Display)]
-#[display("{prefix}({subtype})")]
-pub struct ProprietaryKeyType {
-    pub prefix: String,
-    pub subtype: u8,
-}
-
-impl FromStr for ProprietaryKeyType {
-    type Err = ProprietaryKeyError;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        let mut split = s.trim_end_matches(')').split('(');
-        match (
-            split.next().map(str::to_owned),
-            split.next().map(u8::from_str).transpose().ok().flatten(),
-            split.next(),
-        ) {
-            (Some(prefix), Some(subtype), None) => Ok(ProprietaryKeyType { prefix, subtype }),
-            _ => Err(ProprietaryKeyError::WrongType(s.to_owned())),
-        }
-    }
-}
-
-// --proprietary-key "input(1) DBC(1) 8536ba03:~"
-#[derive(Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Debug)]
-pub struct ProprietaryKeyDescriptor {
-    pub location: ProprietaryKeyLocation,
-    pub ty: ProprietaryKeyType,
-    pub key: Option<Vec<u8>>,
-    pub value: Option<Vec<u8>>,
-}
-
-impl From<ProprietaryKeyDescriptor> for ProprietaryKey {
-    fn from(key: ProprietaryKeyDescriptor) -> Self { ProprietaryKey::from(&key) }
-}
-
-impl From<&ProprietaryKeyDescriptor> for ProprietaryKey {
-    fn from(key: &ProprietaryKeyDescriptor) -> Self {
-        ProprietaryKey {
-            prefix: key.ty.prefix.as_bytes().to_vec(),
-            subtype: key.ty.subtype,
-            key: key.key.as_ref().cloned().unwrap_or_default(),
-        }
-    }
-}
-
-impl FromStr for ProprietaryKeyDescriptor {
-    type Err = ProprietaryKeyError;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        let err = ProprietaryKeyError::WrongFormat(s.to_owned());
-        let mut split = s.split_whitespace();
-        match (
-            split
-                .next()
-                .map(ProprietaryKeyLocation::from_str)
-                .transpose()?,
-            split.next().map(ProprietaryKeyType::from_str).transpose()?,
-            split.next().and_then(|s| s.split_once(':')),
-            split.next(),
-        ) {
-            (Some(location), Some(ty), None, None) => Ok(ProprietaryKeyDescriptor {
-                location,
-                ty,
-                key: None,
-                value: None,
-            }),
-            (Some(location), Some(ty), Some((k, v)), None) => Ok(ProprietaryKeyDescriptor {
-                location,
-                ty,
-                key: if k.is_empty() {
-                    None
-                } else {
-                    Some(Vec::from_hex(k).map_err(|_| err.clone())?)
-                },
-                value: if v.is_empty() {
-                    None
-                } else {
-                    Some(Vec::from_hex(v).map_err(|_| err)?)
-                },
-            }),
-            _ => Err(err),
-        }
-    }
-}
-
-impl Display for ProprietaryKeyDescriptor {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        write!(f, "{} {} ", self.location, self.ty)?;
-        if (&self.key, &self.value) == (&None, &None) {
-            return Ok(());
-        }
-        write!(
-            f,
-            "{}:{}",
-            self.key.as_deref().map(<[u8]>::to_hex).unwrap_or_default(),
-            self.value
-                .as_deref()
-                .map(<[u8]>::to_hex)
-                .unwrap_or_default()
-        )
-    }
-}
-
 #[derive(Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Debug, Display, From)]
 #[display(inner)]
 #[allow(clippy::large_enum_variant)]
@@ -1064,7 +920,7 @@ pub enum Error {
     StrictEncoding(strict_encoding::Error),
 
     #[from]
-    ConsensusEncoding(bitcoin::consensus::encode::Error),
+    PsbtEncoding(consensus::encode::Error),
 
     #[from]
     Miniscript(miniscript::Error),
@@ -1098,6 +954,9 @@ pub enum Error {
     #[display(doc_comments)]
     DescriptorDerivePattern,
 
+    /// allowing tapret commitments with `--`
+    TapretRequiresTaproot,
+
     /// error in extended key encoding: {0}
     #[from]
     #[display(doc_comments)]
